from dune.perftool.generation import temporary_variable, instruction
from dune.perftool.options import get_option
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.argument import name_accumulation_variable
from dune.perftool.pdelab.localoperator import boundary_predicates
import pymbolic.primitives as prim


def define_accumulation_alias(container, accumspace):
    k = get_option("number_of_blocks")
    p = accumspace.element.degree()
    temporary_variable(container+"_alias", shape=(k + 1, k + 1, p + 1, p + 1), strides=(p * k + 1, p, k + 1, 1),
                       base_storage="dummy_"+container, managed=True, _base_storage_access_may_be_aliasing=True)
    code = "{} = &{}.container()({},0);".format(container + "_alias", container, accumspace.lfs.name)
    instruction(within_inames=frozenset(),
                code=code,
                read_variables=frozenset({container}),
                assignees=frozenset({container+"_alias"}))


def name_accumulation_alias(accumvar, accumspace):
    name = accumvar+"_alias"
    define_accumulation_alias(accumvar, accumspace)
    return name


def generate_accumulation_instruction(expr, visitor):
    # Collect the lfs and lfs indices for the accumulate call
    test_lfs = determine_accumulation_space(visitor.test_info, 0)
    # In the jacobian case, also determine the space for the ansatz space
    ansatz_lfs = determine_accumulation_space(visitor.trial_info, 1)

    # Collect the lfs and lfs indices for the accumulate call
    accumvar = name_accumulation_variable(test_lfs.get_restriction() + ansatz_lfs.get_restriction())
    accumvar_alias = name_accumulation_alias(accumvar, test_lfs)

    predicates = boundary_predicates(expr, visitor.measure, visitor.subdomain_id)

    quad_inames = visitor.interface.quadrature_inames()
    lfs_inames = visitor.test_info.inames
    if visitor.trial_info:
        lfs_inames = lfs_inames + visitor.trial_info.inames

    assignee = prim.Subscript(prim.Variable(accumvar_alias), tuple(prim.Variable(i) for i in lfs_inames))

    instruction(assignee=assignee,
                expression=prim.Sum((expr, assignee)),
                forced_iname_deps=frozenset(lfs_inames).union(frozenset(quad_inames)),
                forced_iname_deps_is_final=True,
                predicates=predicates
                )
